"""
MIT License

Copyright (c) 2017-present Devoxin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import os
import re
import sys
import traceback
from subprocess import PIPE, Popen
from time import monotonic
from typing import Optional, Sequence

import requests

RELEASES_URL = 'https://api.github.com/repos/lavalink-devs/Lavalink/releases'
APPLICATION_BASE_URL = 'https://raw.githubusercontent.com/lavalink-devs/Lavalink/{}/LavalinkServer/application.yml.example'
SEMVER_REGEX = re.compile(r'(\d+)\.(\d+)(?:\.(\d+))?(?:-(\w+)(?:\.(\d+)))?')


class Release:
    def __init__(self, release_json):
        self.tag: str = release_json['tag_name']
        self.major_version: int = int(self.tag[0])
        self.prerelease: bool = release_json['prerelease']
        self.draft: bool = release_json['draft']

        assets = release_json['assets']
        jars = [asset['browser_download_url'] for asset in assets if asset['name'].endswith('.jar')]
        self.download_url: Optional[str] = jars[0] if jars else None

    def __str__(self) -> str:
        return f'{self.tag} {"[prerelease]" if self.prerelease else ""}'

    def __eq__(self, other):
        if not isinstance(other, Release):
            return False

        this_match = SEMVER_REGEX.match(self.tag)
        other_match = SEMVER_REGEX.match(other.tag)

        if not this_match or not other_match:
            raise ValueError('Cannot compare version strings as they do not match the regex pattern')

        this_major, this_minor, this_patch, _, this_build = this_match.groups()
        this_version = (int(this_major), int(this_minor), int(this_patch or 0), int(this_build or 0))

        other_major, other_minor, other_patch, _, other_build = other_match.groups()
        other_version = (int(other_major), int(other_minor), int(other_patch or 0), int(other_build or 0))

        return this_version == other_version

    def __lt__(self, other):
        this_match = SEMVER_REGEX.match(self.tag)

        if not this_match:
            raise ValueError('Cannot compare version strings as they do not match the regex pattern')

        this_major, this_minor, this_patch, _, this_build = this_match.groups()
        this_version = (int(this_major), int(this_minor), int(this_patch or 0), int(this_build or 0))

        if isinstance(other, str):
            parts = list(map(int, other.split('.')))

            if len(parts) == 1:
                other_version = (parts[0], 0, 0, 0)
            elif len(parts) == 2:
                other_version = (parts[0], parts[1], 0, 0)
            elif len(parts) == 3:
                other_version = (parts[0], parts[1], parts[2], 0)
            else:
                raise ValueError('Cannot compare version string with more fields than major.minor.patch')
        elif isinstance(other, Release):
            other_match = SEMVER_REGEX.match(other.tag)

            if not this_match or not other_match:
                raise ValueError('Cannot compare version strings as they do not match the regex pattern')

            other_major, other_minor, other_patch, _, other_build = other_match.groups()
            other_version = (int(other_major), int(other_minor), int(other_patch or 0), int(other_build or 0))
        else:
            raise TypeError(f'"<" not supported between instances of "{type(self).__name__}" and "{type(other).__name__}"')

        return this_version < other_version

    def __gt__(self, other):
        this_match = SEMVER_REGEX.match(self.tag)

        if not this_match:
            raise ValueError('Cannot compare version strings as they do not match the regex pattern')

        this_major, this_minor, this_patch, _, this_build = this_match.groups()
        this_version = (int(this_major), int(this_minor), int(this_patch or 0), int(this_build or 0))

        if isinstance(other, str):
            parts = list(map(int, other.split('.')))

            if len(parts) == 1:
                other_version = (parts[0], 0, 0, 0)
            elif len(parts) == 2:
                other_version = (parts[0], parts[1], 0, 0)
            elif len(parts) == 3:
                other_version = (parts[0], parts[1], parts[2], 0)
            else:
                raise ValueError('Cannot compare version string with more fields than major.minor.patch')
        elif isinstance(other, Release):
            other_match = SEMVER_REGEX.match(other.tag)

            if not this_match or not other_match:
                raise ValueError('Cannot compare version strings as they do not match the regex pattern')

            other_major, other_minor, other_patch, _, other_build = other_match.groups()
            other_version = (int(other_major), int(other_minor), int(other_patch or 0), int(other_build or 0))
        else:
            raise TypeError(f'">" not supported between instances of "{type(self).__name__}" and "{type(other).__name__}"')

        return this_version > other_version

    def __ge__(self, other):
        return self.__eq__(other) or self.__gt__(other)

    def __le__(self, other):
        return self.__eq__(other) or self.__lt__(other)


def display_help():
    print("""
download - Find and download specific Lavalink server versions.
    --no-overwrite  Renames an existing lavalink.jar to lavalink.old.jar
config   - Downloads a fresh application.yml.
    --fetch-dev     Fetches the latest application.yml from the development branch.
    --no-overwrite  Renames an existing application.yml to application.old.yml.
info     - Extracts version and build information from an existing Lavalink.jar.
    """.strip(), file=sys.stdout)


def format_bytes(length: float) -> str:
    sizes = ['B', 'KB', 'MB', 'GB', 'TB']
    unit = 0

    while length >= 1024 and unit < len(sizes) - 1:
        unit += 1
        length /= 1024

    return f'{length:.2f} {sizes[unit]}'


def download(dl_url, path):
    res = requests.get(dl_url, stream=True, timeout=15)

    download_begin = round(monotonic() * 1000)

    def report_progress(cur, tot):
        bar_len = 32
        progress = float(cur) / tot
        filled_len = int(round(bar_len * progress))
        percent = round(progress * 100, 2)
        elapsed = round(monotonic() * 1000) - download_begin

        if elapsed > 0:
            correction = 1000 / elapsed
            speed = cur * correction
        else:
            speed = 0  # placeholder until we have enough data to calculate

        progress_bar = '█' * filled_len + ' ' * (bar_len - filled_len)
        sys.stdout.write(f'Downloading |{progress_bar}| {percent:.1f}% ({cur}/{tot}, {format_bytes(speed)}/s)\r')
        sys.stdout.flush()

        if cur >= tot:
            sys.stdout.write('\n')

    def read_chunk(out, chunk_size=8192):
        total_bytes = int(res.headers['Content-Length'].strip())
        current_bytes = 0

        for chunk in res.iter_content(chunk_size):
            out.write(chunk)
            current_bytes += len(chunk)
            report_progress(min(current_bytes, total_bytes), total_bytes)

    with open(path, 'wb') as out:
        read_chunk(out)


def select_release_unattended(non_draft: Sequence[Release], version_selector: str) -> Release:
    matcher = SEMVER_REGEX.match(version_selector)

    if matcher:
        def exact_version(release: Release):
            return release.tag == version_selector

        predicate = exact_version
    elif version_selector.startswith('>='):
        def gte(release: Release):
            return release >= version_selector[2:]

        predicate = gte
    elif version_selector.startswith('<='):
        def lte(release: Release):
            return release.tag <= version_selector[2:]

        predicate = lte
    elif version_selector.startswith('>'):
        def gt(release: Release):  # pylint: disable=C0103
            return release > version_selector[1:]

        predicate = gt
    elif version_selector.startswith('<'):
        def lte(release: Release):
            return release.tag < version_selector[1:]

        predicate = lte
    elif version_selector.startswith('~='):
        minimum = version_selector[2:]
        major, minor, _ = minimum.split('.')
        maximum = f'{major}.{int(minor) + 1}.0'

        def compatible(release: Release):
            return minimum <= release < maximum

        predicate = compatible
    else:
        # TODO: Support multiple version specifiers (e.g. >=3.7.0,<4.0.0)
        raise ValueError('Unsupported version selector')

    selected_release = next((release for release in non_draft if predicate(release)), None)

    if not selected_release:
        print('Couldn\'t find a suitable release with the provided version selector.', file=sys.stderr)
        sys.exit(1)

    print(f'Release selected: {selected_release.tag}', file=sys.stdout)

    return selected_release


def select_release(non_draft: Sequence[Release]) -> Release:
    suitable_releases = []

    for release in non_draft:
        if not release.download_url:
            continue

        newest: Optional[Release] = next((sr for sr in suitable_releases if sr.major_version == release.major_version), None)

        if newest:
            if newest > release:  # GitHub gives newest->oldest releases, so it could be that we iterate over a pre-release before a release.
                if newest.prerelease and not release.prerelease:  # If that is the case, we check the version against the current non-prerelease
                    current_non_prerelease: Optional[Release] = next((sr for sr in suitable_releases if sr.major_version == release.major_version
                                                                      and not sr.prerelease), None)

                    if current_non_prerelease and current_non_prerelease > release:
                        continue
                else:
                    continue

        suitable_releases.append(release)

    if not suitable_releases:
        print('No suitable Lavalink releases were found.', file=sys.stdout)
        sys.exit(0)  # Perhaps this should be an error, however this could also be valid (but very unlikely).

    if len(suitable_releases) > 1:
        print('There are multiple Lavalink versions to choose from.\n'
              'They have automatically been filtered based on their version, and whether they are a pre-release.\n\n'
              'Type the number of the release you would like to download.\n', file=sys.stdout)

        for index, release in enumerate(suitable_releases, start=1):
            print(f'[{index}] {release}', file=sys.stdout)

        try:
            selected = int(input('> ')) - 1

            if not 0 <= selected <= len(suitable_releases):
                raise ValueError
        except ValueError:
            print('An incorrect selection has been made, cancelling...', file=sys.stderr)
            sys.exit(1)
    else:
        selected = 0

    return suitable_releases[selected]


def download_jar(arguments: Sequence[str]):
    try:
        res = requests.get(RELEASES_URL, timeout=15).json()
    except requests.exceptions.JSONDecodeError:
        print('Failed to retrieve Lavalink releases', file=sys.stderr)
        sys.exit(1)

    releases = list(map(Release, res))
    non_draft = [r for r in releases if not r.draft]

    if arguments:
        try:
            release = select_release_unattended(non_draft, arguments[0])
        except ValueError:
            traceback.print_exc(file=sys.stderr)
            sys.exit(1)
    else:
        release = select_release(non_draft)

    cwd = os.getcwd()
    dl_url = release.download_url
    dl_path = os.path.join(cwd, 'lavalink.jar')

    if '--no-overwrite' in arguments and os.path.exists(dl_path):
        os.rename(dl_path, os.path.join(cwd, 'lavalink.old.jar'))

    download(dl_url, dl_path)
    print(f'Downloaded {release.tag} to {dl_path}', file=sys.stdout)
    sys.exit(0)


def download_config(arguments: Sequence[str], branch: str):
    cwd = os.getcwd()
    dl_url = APPLICATION_BASE_URL.format(branch)
    dl_path = os.path.join(cwd, 'application.yml')

    if '--no-overwrite' in arguments and os.path.exists(dl_path):
        os.rename(dl_path, os.path.join(cwd, 'application.old.yml'))

    download(dl_url, dl_path)
    print(f'Downloaded to {dl_path}', file=sys.stdout)
    sys.exit(0)


def print_info(arguments: Sequence[str]):
    check_names = ['lavalink.jar', 'Lavalink.jar', 'LAVALINK.JAR']

    if arguments:
        check_names.extend([arguments[0]])

    file_name = next((fn for fn in check_names if os.path.exists(fn)), None)

    if not file_name:
        print('Unable to display Lavalink server info: No Lavalink file found.', file=sys.stderr)
        sys.exit(1)

    proc = Popen(['java', '-jar', file_name, '--version'], stdout=PIPE, stderr=PIPE, text=True)
    stdout, stderr = proc.communicate()

    if stderr:
        if 'UnsupportedClassVersionError' in stderr:
            java_proc = Popen(['java', '-version'], stdout=PIPE, stderr=PIPE, text=True)
            j_stdout, j_stderr = java_proc.communicate()
            j_ver = re.search(r'java version "([\d._]*)"', j_stdout or j_stderr)
            java_version = j_ver.group(1) if j_ver else 'UNKNOWN'

            if java_version.startswith('1.8'):
                java_version = f'8/{java_version}'

            print(f'Unable to display Lavalink server info.\nYour Java version is out of date. (Java {java_version})\n\n'
                  'Java 11+ is required to run Lavalink.', file=sys.stderr)
            sys.exit(1)

        print(stderr, file=sys.stderr)
        sys.exit(1)
    else:
        print(stdout.strip(), file=sys.stdout)
        sys.exit(0)


def main():
    if len(sys.argv) < 2 or sys.argv[1] in ('--help', 'help', '?', '/help'):
        display_help()
        return

    _, action, *arguments = sys.argv
    target_branch = 'dev' if '--fetch-dev' in arguments else 'master'

    try:
        if action == 'download':
            download_jar(arguments)
        elif action == 'config':
            download_config(arguments, target_branch)
        elif action == 'info':
            print_info(arguments)
        else:
            print(f'Invalid argument \'{action}\'. Use --help to show usage.', file=sys.stderr)
            sys.exit(1)
    except KeyboardInterrupt:
        sys.exit(2)  # CTRL-C = SIGINT = 2


if __name__ == '__main__':
    main()



"""
MIT License

Copyright (c) 2017-present Devoxin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from typing import TYPE_CHECKING, Final

if TYPE_CHECKING:
    from .node import Node

__all__ = (
    "Penalty",
    "Stats",
)


class Penalty:
    """
    Represents the penalty of the stats of a Node.

    Attributes
    ----------
    player_penalty: :class:`int`
        The number of playing players. 1 player = 1 penalty point.
    cpu_penalty: :class:`float`
        The penalty incurred from system CPU usage.
    null_frame_penalty: :class:`float`
        The penalty incurred from the average number of null frames per minute.
    deficit_frame_penalty: :class:`float`
        The penalty incurred from the average number of deficit frames per minute.
    total: :class:`float`
        The overall penalty, as a sum of all other penalties.
    """
    __slots__ = ('player_penalty', 'cpu_penalty', 'null_frame_penalty', 'deficit_frame_penalty', 'total')

    def __init__(self, stats):
        self.player_penalty: Final[int] = stats.playing_players
        self.cpu_penalty: Final[float] = 1.05 ** (100 * stats.system_load) * 10 - 10
        self.null_frame_penalty: Final[float] = ((1.03 ** (500 * (stats.frames_nulled / 3000))) * 300 - 300) * 2
        self.deficit_frame_penalty: Final[float] = (1.03 ** (500 * (stats.frames_deficit / 3000))) * 600 - 600
        self.total: Final[float] = self.player_penalty + self.cpu_penalty + self.null_frame_penalty + self.deficit_frame_penalty


class Stats:
    """
    Encapsulates the 'Statistics' emitted by Lavalink, usually every minute.

    Attributes
    ----------
    is_fake: :class:`bool`
        Whether or not the stats are accurate. This should only be True when
        the node has not yet received any statistics from the Lavalink server.
    uptime: :class:`int`
        How long the node has been running for, in milliseconds.
    players: :class:`int`
        The number of players connected to the node.
    playing_players: :class:`int`
        The number of players that are playing in the node.
    memory_free: :class:`int`
        The amount of memory free to the node, in bytes.
    memory_used: :class:`int`
        The amount of memory that is used by the node, in bytes.
    memory_allocated: :class:`int`
        The amount of memory allocated to the node, in bytes.
    memory_reservable: :class:`int`
        The amount of memory reservable to the node, in bytes.
    cpu_cores: :class:`int`
        The amount of cpu cores the system of the node has.
    system_load: :class:`float`
        The overall CPU load of the system. This is a number between 0-1,
        but can be multiplied by 100 for the percentage (0-100).
    lavalink_load: :class:`float`
        The CPU load generated by Lavalink This is a number between 0-1,
        but can be multiplied by 100 for the percentage (0-100).
    frames_sent: :class:`int`
        The number of frames sent to Discord.

        Warning
        -------
        Given that audio packets are sent via UDP, this number may not be 100% accurate due to packets dropped in transit.
    frames_nulled: :class:`int`
        The number of frames that yielded null, rather than actual data.
    frames_deficit: :class:`int`
        The number of missing frames. Lavalink generates this figure by calculating how many packets to expect
        per minute, and deducting ``frames_sent``. Deficit frames could mean the CPU is overloaded, and isn't
        generating frames as quickly as it should be.
    penalty: :class:`Penalty`
    """
    __slots__ = ('_node', 'is_fake', 'uptime', 'players', 'playing_players', 'memory_free', 'memory_used', 'memory_allocated',
                 'memory_reservable', 'cpu_cores', 'system_load', 'lavalink_load', 'frames_sent', 'frames_nulled',
                 'frames_deficit', 'penalty')

    def __init__(self, node, data):
        self._node: Final['Node'] = node

        self.is_fake: Final[bool] = data.get('isFake', False)
        self.uptime: Final[int] = data['uptime']

        self.players: Final[int] = data['players']
        self.playing_players: Final[int] = data['playingPlayers']

        memory = data['memory']
        self.memory_free: Final[int] = memory['free']
        self.memory_used: Final[int] = memory['used']
        self.memory_allocated: Final[int] = memory['allocated']
        self.memory_reservable: Final[int] = memory['reservable']

        cpu = data['cpu']
        self.cpu_cores: Final[int] = cpu['cores']
        self.system_load: Final[float] = cpu['systemLoad']
        self.lavalink_load: Final[float] = cpu['lavalinkLoad']

        frame_stats = data.get('frameStats') or {}
        self.frames_sent: Final[int] = frame_stats.get('sent', 0)
        self.frames_nulled: Final[int] = frame_stats.get('nulled', 0)
        self.frames_deficit: Final[int] = frame_stats.get('deficit', 0)
        self.penalty: Final[Penalty] = Penalty(self)

    @classmethod
    def empty(cls, node):
        data = {
            'isFake': True,
            'uptime': 0,
            'players': 0,
            'playingPlayers': 0,
            'memory': {
                'free': 0,
                'used': 0,
                'allocated': 0,
                'reservable': 0
            },
            'cpu': {
                'cores': 0,
                'systemLoad': 0,
                'lavalinkLoad': 0
            }
        }

        return cls(node, data)



# flake8: noqa

__title__ = 'Lavalink'
__author__ = 'Devoxin'
__license__ = 'MIT'
__copyright__ = 'Copyright 2017-present Devoxin'
__version__ = '5.11.0'


from typing import Type

from .abc import *
from .client import *
from .dataio import *
from .errors import *
from .events import *
from .filters import *
from .node import *
from .nodemanager import *
from .player import *
from .playermanager import *
from .server import *
from .source_decoders import *
from .stats import *
from .utils import *


def listener(*events: Type[Event]):
    """
    Marks this function as an event listener for Lavalink.py.
    This **must** be used on class methods, and you must ensure that you register
    decorated methods by using :func:`Client.add_event_hooks`.

    Example:

        .. code:: python

            @listener()
            async def on_lavalink_event(self, event):  # Event can be ANY Lavalink event
                ...

            @listener(TrackStartEvent)
            async def on_track_start(self, event: TrackStartEvent):
                ...

    Note
    ----
    Track event dispatch order is not guaranteed!
    For example, this means you could receive a :class:`TrackStartEvent` before you receive a
    :class:`TrackEndEvent` when executing operations such as ``skip()``.

    Parameters
    ----------
    events: :class:`Event`
        The events to listen for. Leave this empty to listen for all events.
    """
    def wrapper(func):
        setattr(func, '_lavalink_events', events)
        return func
    return wrapper



"""
MIT License

Copyright (c) 2017-present Devoxin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from typing import TYPE_CHECKING, Any, Dict, Final, Optional

if TYPE_CHECKING:
    from .node import Node

__all__ = (
    "LavalinkError",
    "ClientError",
    "AuthenticationError",
    "InvalidTrack",
    "LoadError",
    "RequestError",
)


class LavalinkError(Exception):
    """ Base exception for all errors raised by Lavalink.py. """


class ClientError(LavalinkError):
    """ Raised when something goes wrong within the client. """


class AuthenticationError(LavalinkError):
    """ Raised when a request fails due to invalid authentication. """


class InvalidTrack(LavalinkError):
    """ Raised when an invalid track was passed. """


class LoadError(LavalinkError):
    """ Raised when a track fails to load. E.g. if a DeferredAudioTrack fails to find an equivalent. """


class RequestError(LavalinkError):
    """
    Raised when a request to the Lavalink server fails.

    Attributes
    ----------
    status: :class:`int`
        The HTTP status code returned by the server.
    timestamp: :class:`int`
        The epoch timestamp in milliseconds, at which the error occurred.
    error: :class:`str`
        The HTTP status code message.
    message: :class:`str`
        The error message.
    path: :class:`str`
        The request path.
    trace: Optional[:class:`str`]
        The stack trace of the error. This will only be present if ``trace=true`` was provided
        in the query parameters of the request.
    params: Dict[str, Any]
        The parameters passed to the request that errored.
    """
    __slots__ = ('node', 'status', 'timestamp', 'error', 'message', 'path', 'trace', 'params')

    def __init__(self, node: 'Node', status: int, response: dict, params: Dict[str, Any]):
        self.node: Final['Node'] = node
        self.status: Final[int] = status
        self.timestamp: Final[int] = response['timestamp']
        self.error: Final[str] = response['error']
        self.message: Final[str] = response.get('message', '')
        self.path: Final[str] = response['path']
        self.trace: Final[Optional[str]] = response.get('trace', None)
        self.params: Final[Dict[str, Any]] = params
        super().__init__(f'An error occurred during a request to the node ({self.message}): {self.error}')



"""
MIT License

Copyright (c) 2017-present Devoxin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
# Necessary evil due to documenting filter update kwargs.
# At least, until I can come up with a better solution to doing this.
# pylint: disable=arguments-differ

from typing import Any, Dict, Sequence, Tuple, overload

from .abc import Filter
from .common import MISSING

__all__ = (
    "Volume",
    "Equalizer",
    "Karaoke",
    "Timescale",
    "Tremolo",
    "Vibrato",
    "Rotation",
    "LowPass",
    "ChannelMix",
    "Distortion",
)


class Volume(Filter[float]):
    """
    Adjusts the audio output volume.
    """
    def __init__(self, volume: float = 1.0):
        super().__init__(volume)

    def update(self, *, volume: float):  # type: ignore
        """
        Modifies the player volume.
        This uses LavaDSP's volume filter, rather than Lavaplayer's native
        volume changer.

        Note
        ----
        The limits are:

            0 ≤ volume ≤ 5

        Parameters
        ----------
        volume: :class:`float`
            The new volume of the player. 1.0 means 100%/default.
        """
        volume = float(volume)

        if not 0 <= volume <= 5:
            raise ValueError('volume must be bigger than or equal to 0, and less than or equal to 5.')

        self.values = volume

    def serialize(self) -> Dict[str, float]:
        return {'volume': self.values}


class Equalizer(Filter[Sequence[float]]):
    """
    Allows modifying the gain of 15 bands, to boost or reduce the volume of specific frequency ranges.
    For example, this could be used to boost the low (bass) frequencies to act as a 'bass boost'.
    """
    def __init__(self, gains: Sequence[float] = MISSING):
        super().__init__([0.0] * 15 if gains is MISSING else gains)

    @overload
    def update(self, *, bands: Sequence[Tuple[int, float]]):
        ...

    @overload
    def update(self, *, band: int, gain: float):
        ...

    def update(self, **kwargs):
        """
        Modifies the gain of each specified band.
        There are 15 total bands (indexes 0 to 14) that can be modified.
        The meaningful range of each band is -0.25 (muted) to 1.0. A gain of 0.25 doubles the frequency.
        The default gain is 0.0.
        Modifying the gain could alter the volume of output.

        The frequencies of each band are as follows:
        25 Hz, 40 Hz, 63 Hz, 100 Hz, 160 Hz, 250 Hz, 400 Hz, 630 Hz, 1k Hz, 1.6k Hz, 2.5k Hz, 4k Hz, 6.3k Hz, 10k Hz, 16k Hz
        Leftmost frequency represents band 0, rightmost frequency represents band 14.

        Note
        ----
        You can provide either ``bands`` OR ``band`` and ``gain`` for the parameters.

        The limits are:

            0 ≤ band ≤ 14

            -0.25 ≤ gain ≤ 1.0

        Parameters
        ----------
        bands: Sequence[Tuple[:class:`int`, :class:`float`]]
            The bands to modify, and their respective gains.
        band: :class:`int`
            The band to modify.
        gain: :class:`float`
            The new gain of the band.

        Raises
        ------
        :class:`ValueError`
        """
        if 'bands' in kwargs:
            bands = kwargs.pop('bands')

            sanity_check = isinstance(bands, list) and all(isinstance(pair, tuple) for pair in bands) and \
                all(isinstance(band, int) and isinstance(gain, (float, int)) for band, gain in bands) and \
                all(0 <= band <= 14 and -0.25 <= gain <= 1.0 for band, gain in bands)

            if not sanity_check:
                raise ValueError('Bands must be a list of tuple representing (band: int, gain: float) with values between '
                                 '0 to 14, and -0.25 to 1.0 respectively')

            for band, gain in bands:
                self.values[band] = gain
        elif 'band' in kwargs and 'gain' in kwargs:
            band = int(kwargs.pop('band'))
            gain = float(kwargs.pop('gain'))

            if not 0 <= band <= 14:
                raise ValueError('Band must be between 0 and 14 (start and end inclusive)')

            if not -0.25 <= gain <= 1.0:
                raise ValueError('Gain must be between -0.25 and 1.0 (start and end inclusive)')

            self.values[band] = gain
        else:
            raise KeyError('Expected parameter bands OR band and gain, but neither were provided')

    def serialize(self) -> Dict[str, Any]:
        return {'equalizer': [{'band': band, 'gain': gain} for band, gain in enumerate(self.values)]}


class Karaoke(Filter[Dict[str, float]]):
    """
    Allows for isolating a frequency range (commonly, the vocal range).
    Useful for karaoke/sing-along.
    """
    def __init__(self, level: float = 1.0, mono_level: float = 1.0,
                 filter_band: float = 220.0, filter_width: float = 100.0):
        super().__init__({'level': level, 'monoLevel': mono_level, 'filterBand': filter_band, 'filterWidth': filter_width})

    @overload
    def update(self, *, level: float):
        ...

    @overload
    def update(self, *, mono_level: float):
        ...

    @overload
    def update(self, *, filter_band: float):
        ...

    @overload
    def update(self, *, filter_width: float):
        ...

    @overload
    def update(self, *, level: float, mono_level: float):
        ...

    @overload
    def update(self, *, level: float, filter_width: float):
        ...

    @overload
    def update(self, *, level: float, filter_band: float):
        ...

    @overload
    def update(self, *, mono_level: float, filter_width: float):
        ...

    @overload
    def update(self, *, mono_level: float, filter_band: float):
        ...

    @overload
    def update(self, *, filter_band: float, filter_width: float):
        ...

    @overload
    def update(self, *, level: float, mono_level: float, filter_width: float):
        ...

    @overload
    def update(self, *, level: float, mono_level: float, filter_band: float):
        ...

    @overload
    def update(self, *, level: float, mono_level: float, filter_band: float, filter_width: float):
        ...

    def update(self, **kwargs):
        """
        Parameters
        ----------
        level: :class:`float`
            The level of the Karaoke effect.
        mono_level: :class:`float`
            The mono level of the Karaoke effect.
        filter_band: :class:`float`
            The frequency of the band to filter.
        filter_width: :class:`float`
            The width of the filter.

        Raises
        ------
        :class:`ValueError`
        """
        if 'level' in kwargs:
            self.values['level'] = float(kwargs.pop('level'))

        if 'mono_level' in kwargs:
            self.values['monoLevel'] = float(kwargs.pop('mono_level'))

        if 'filter_band' in kwargs:
            self.values['filterBand'] = float(kwargs.pop('filter_band'))

        if 'filter_width' in kwargs:
            self.values['filterWidth'] = float(kwargs.pop('filter_width'))

    def serialize(self) -> Dict[str, Dict[str, float]]:
        return {'karaoke': self.values}


class Timescale(Filter[Dict[str, float]]):
    """
    Allows speeding up/slowing down the audio, adjusting the pitch and playback rate.
    """
    def __init__(self, speed: float = 1.0, pitch: float = 1.0, rate: float = 1.0):
        super().__init__({'speed': speed, 'pitch': pitch, 'rate': rate})

    @overload
    def update(self, *, speed: float):
        ...

    @overload
    def update(self, *, pitch: float):
        ...

    @overload
    def update(self, *, rate: float):
        ...

    @overload
    def update(self, *, speed: float, pitch: float):
        ...

    @overload
    def update(self, *, speed: float, rate: float):
        ...

    @overload
    def update(self, *, rate: float, pitch: float):
        ...

    @overload
    def update(self, *, speed: float, rate: float, pitch: float):
        ...

    def update(self, **kwargs):
        """
        Note
        ----
        The limits are:

            0.1 ≤ speed

            0.1 ≤ pitch

            0.1 ≤ rate

        Parameters
        ----------
        speed: :class:`float`
            The playback speed.
        pitch: :class:`float`
            The pitch of the audio.
        rate: :class:`float`
            The playback rate.

        Raises
        ------
        :class:`ValueError`
        """
        if 'speed' in kwargs:
            speed = float(kwargs.pop('speed'))

            if speed <= 0:
                raise ValueError('Speed must be bigger than 0')

            self.values['speed'] = speed

        if 'pitch' in kwargs:
            pitch = float(kwargs.pop('pitch'))

            if pitch <= 0:
                raise ValueError('Pitch must be bigger than 0')

            self.values['pitch'] = pitch

        if 'rate' in kwargs:
            rate = float(kwargs.pop('rate'))

            if rate <= 0:
                raise ValueError('Rate must be bigger than 0')

            self.values['rate'] = rate

    def serialize(self) -> Dict[str, Dict[str, float]]:
        return {'timescale': self.values}


class Tremolo(Filter[Dict[str, float]]):
    """
    Applies a 'tremble' effect to the audio.
    """
    def __init__(self, frequency: float = 2.0, depth: float = 0.5):
        super().__init__({'frequency': frequency, 'depth': depth})

    @overload
    def update(self, *, frequency: float):
        ...

    @overload
    def update(self, *, depth: float):
        ...

    @overload
    def update(self, *, frequency: float, depth: float):
        ...

    def update(self, **kwargs):
        """
        Note
        ----
        The limits are:

            0 < frequency

            0 < depth ≤ 1

        Parameters
        ----------
        frequency: :class:`float`
            How frequently the effect should occur.
        depth: :class:`float`
            The "strength" of the effect.

        Raises
        ------
        :class:`ValueError`
        """
        if 'frequency' in kwargs:
            frequency = float(kwargs.pop('frequency'))

            if frequency < 0:
                raise ValueError('Frequency must be bigger than 0')

            self.values['frequency'] = frequency

        if 'depth' in kwargs:
            depth = float(kwargs.pop('depth'))

            if not 0 < depth <= 1:
                raise ValueError('Depth must be bigger than 0, and less than or equal to 1.')

            self.values['depth'] = depth

    def serialize(self) -> Dict[str, Dict[str, float]]:
        return {'tremolo': self.values}


class Vibrato(Filter[Dict[str, float]]):
    """
    Applies a 'wobble' effect to the audio.
    """
    def __init__(self, frequency: float = 2.0, depth: float = 0.5):
        super().__init__({'frequency': frequency, 'depth': depth})

    @overload
    def update(self, *, frequency: float):
        ...

    @overload
    def update(self, *, depth: float):
        ...

    @overload
    def update(self, *, frequency: float, depth: float):
        ...

    def update(self, **kwargs):
        """
        Note
        ----
        The limits are:

            0 < frequency ≤ 14

            0 < depth ≤ 1

        Parameters
        ----------
        frequency: :class:`float`
            How frequently the effect should occur.
        depth: :class:`float`
            The "strength" of the effect.

        Raises
        ------
        :class:`ValueError`
        """
        if 'frequency' in kwargs:
            frequency = float(kwargs.pop('frequency'))

            if not 0 < frequency <= 14:
                raise ValueError('Frequency must be bigger than 0, and less than or equal to 14')

            self.values['frequency'] = frequency

        if 'depth' in kwargs:
            depth = float(kwargs.pop('depth'))

            if not 0 < depth <= 1:
                raise ValueError('Depth must be bigger than 0, and less than or equal to 1.')

            self.values['depth'] = depth

    def serialize(self) -> Dict[str, Dict[str, float]]:
        return {'vibrato': self.values}


class Rotation(Filter[float]):
    """
    Phases the audio in and out of the left and right channels in an alternating manner.
    This is commonly used to create the 8D effect.
    """
    def __init__(self, rotation_hz: float = 0.0):
        super().__init__(rotation_hz)

    def update(self, *, rotation_hz: float):  # type: ignore
        """
        Note
        ----
        The limits are:

            0 ≤ rotation_hz

        Parameters
        ----------
        rotation_hz: :class:`float`
            How frequently the effect should occur.

        Raises
        ------
        :class:`ValueError`
        """
        rotation_hz = float(rotation_hz)

        if rotation_hz < 0:
            raise ValueError('rotation_hz must be bigger than or equal to 0')

        self.values = rotation_hz

    def serialize(self) -> Dict[str, Dict[str, float]]:  # type: ignore
        return {'rotation': {'rotationHz': self.values}}


class LowPass(Filter[float]):
    """
    Applies a low-pass effect to the audio, whereby only low frequencies can pass,
    effectively cutting off high frequencies meaning more emphasis is put on lower frequencies.
    """
    def __init__(self, smoothing: float = 20.0):
        super().__init__(smoothing)

    def update(self, *, smoothing: float):  # type: ignore
        """
        Note
        ----
        The limits are:

            1 < smoothing

        Parameters
        ----------
        smoothing: :class:`float`
            The strength of the effect.

        Raises
        ------
        :class:`ValueError`
        """
        smoothing = float(smoothing)

        if smoothing <= 1:
            raise ValueError('smoothing must be bigger than 1')

        self.values = smoothing

    def serialize(self) -> Dict[str, Dict[str, float]]:  # type: ignore
        return {'lowPass': {'smoothing': self.values}}


class ChannelMix(Filter[Dict[str, float]]):
    """
    Allows passing the audio from one channel to the other, or isolating individual
    channels.
    """
    def __init__(self, left_to_left: float = 1.0, left_to_right: float = 0.0,
                 right_to_left: float = 0.0, right_to_right: float = 0.0):
        super().__init__({'leftToLeft': left_to_left, 'leftToRight': left_to_right,
                          'rightToLeft': right_to_left, 'rightToRight': right_to_right})

    def update(self, **kwargs):
        """
        Note
        ----
        The limits are:

            0 ≤ leftToLeft ≤ 1.0

            0 ≤ leftToRight ≤ 1.0

            0 ≤ rightToLeft ≤ 1.0

            0 ≤ rightToRight ≤ 1.0

        Parameters
        ----------
        left_to_left: :class:`float`
            The volume level of the audio going from the "Left" channel to the "Left" channel.
        left_to_right: :class:`float`
            The volume level of the audio going from the "Left" channel to the "Right" channel.
        right_to_left: :class:`float`
            The volume level of the audio going from the "Right" channel to the "Left" channel.
        right_to_right: :class:`float`
            The volume level of the audio going from the "Right" channel to the "Left" channel.

        Raises
        ------
        :class:`ValueError`
        """
        if 'left_to_left' in kwargs:
            left_to_left = float(kwargs.pop('left_to_left'))

            if not 0 <= left_to_left <= 1:
                raise ValueError('left_to_left must be bigger than or equal to 0, and less than or equal to 1.')

            self.values['leftToLeft'] = left_to_left

        if 'left_to_right' in kwargs:
            left_to_right = float(kwargs.pop('left_to_right'))

            if not 0 <= left_to_right <= 1:
                raise ValueError('left_to_right must be bigger than or equal to 0, and less than or equal to 1.')

            self.values['leftToRight'] = left_to_right

        if 'right_to_left' in kwargs:
            right_to_left = float(kwargs.pop('right_to_left'))

            if not 0 <= right_to_left <= 1:
                raise ValueError('right_to_left must be bigger than or equal to 0, and less than or equal to 1.')

            self.values['rightToLeft'] = right_to_left

        if 'right_to_right' in kwargs:
            right_to_right = float(kwargs.pop('right_to_right'))

            if not 0 <= right_to_right <= 1:
                raise ValueError('right_to_right must be bigger than or equal to 0, and less than or equal to 1.')

            self.values['rightToRight'] = right_to_right

    def serialize(self) -> Dict[str, Dict[str, float]]:
        return {'channelMix': self.values}


class Distortion(Filter[Dict[str, float]]):
    """
    As the name suggests, this distorts the audio.
    """
    def __init__(self, sin_offset: float = 0.0, sin_scale: float = 1.0, cos_offset: float = 0.0,
                 cos_scale: float = 1.0, tan_offset: float = 0.0, tan_scale: float = 1.0, offset: float = 0.0,
                 scale: float = 1.0):
        super().__init__({'sinOffset': sin_offset, 'sinScale': sin_scale, 'cosOffset': cos_offset, 'cosScale': cos_scale,
                          'tanOffset': tan_offset, 'tanScale': tan_scale, 'offset': offset, 'scale': scale})

    def update(self, **kwargs):
        """
        Parameters
        ----------
        sin_offset: :class:`float`
            The sin offset.
        sin_scale: :class:`float`
            The sin scale.
        cos_offset: :class:`float`
            The sin offset.
        cos_scale: :class:`float`
            The sin scale.
        tan_offset: :class:`float`
            The sin offset.
        tan_scale: :class:`float`
            The sin scale.
        offset: :class:`float`
            The sin offset.
        scale: :class:`float`
            The sin scale.

        Raises
        ------
        :class:`ValueError`
        """
        if 'sin_offset' in kwargs:
            self.values['sinOffset'] = float(kwargs.pop('sin_offset'))

        if 'sin_scale' in kwargs:
            self.values['sinScale'] = float(kwargs.pop('sin_scale'))

        if 'cos_offset' in kwargs:
            self.values['cosOffset'] = float(kwargs.pop('cos_offset'))

        if 'cos_scale' in kwargs:
            self.values['cosScale'] = float(kwargs.pop('cos_scale'))

        if 'tan_offset' in kwargs:
            self.values['tanOffset'] = float(kwargs.pop('tan_offset'))

        if 'tan_scale' in kwargs:
            self.values['tanScale'] = float(kwargs.pop('tan_scale'))

        if 'offset' in kwargs:
            self.values['offset'] = float(kwargs.pop('offset'))

        if 'scale' in kwargs:
            self.values['scale'] = float(kwargs.pop('scale'))

    def serialize(self) -> Dict[str, Dict[str, float]]:
        return {'distortion': self.values}



"""
MIT License

Copyright (c) 2017-present Devoxin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import struct
from base64 import b64encode
from typing import Any, Callable, Dict, Mapping, Optional, Tuple

from .common import MISSING
from .dataio import DataReader, DataWriter
from .errors import InvalidTrack
from .player import AudioTrack
from .source_decoders import DEFAULT_DECODER_MAPPING

__all__ = (
    "V2_KEYSET",
    "V3_KEYSET",
    "timestamp_to_millis",
    "format_time",
    "parse_time",
    "decode_track",
    "encode_track",
    "encode_track_v2",
    "encode_track_v3",
)

V2_KEYSET = {'title', 'author', 'length', 'identifier', 'isStream', 'uri', 'sourceName', 'position'}
V3_KEYSET = V2_KEYSET | {'artworkUrl', 'isrc'}


def timestamp_to_millis(timestamp: str) -> int:
    """
    Converts a timestamp such as 03:28 or 02:15:53 to milliseconds.

    Example
    -------
    .. code:: python

        await player.play(track, start_time=timestamp_to_millis('01:13'))

    Parameters
    ----------
    timestamp: :class:`str`
        The timestamp to convert into milliseconds.

    Returns
    -------
    :class:`int`
        The millisecond value of the timestamp.
    """
    try:
        sections = list(map(int, timestamp.split(':')))
    except ValueError as error:
        raise ValueError('Timestamp should consist of integers and colons only') from error

    if not sections:
        raise TypeError('An invalid timestamp was provided, a timestamp should look like 1:30')

    if len(sections) > 4:
        raise TypeError('Too many segments within the provided timestamp! Provide no more than 4 segments.')

    if len(sections) == 4:
        days, hours, minutes, seconds = map(int, sections)
        return (days * 86400000) + (hours * 3600000) + (minutes * 60000) + (seconds * 1000)

    if len(sections) == 3:
        hours, minutes, seconds = map(int, sections)
        return (hours * 3600000) + (minutes * 60000) + (seconds * 1000)

    if len(sections) == 2:
        minutes, seconds = map(int, sections)
        return (minutes * 60000) + (seconds * 1000)

    seconds, = map(int, sections)
    return seconds * 1000


def format_time(time: int) -> str:
    """
    Formats the given time into HH:MM:SS.

    Parameters
    ----------
    time: :class:`int`
        The time in milliseconds.

    Returns
    -------
    :class:`str`
    """
    hours, remainder = divmod(time / 1000, 3600)
    minutes, seconds = divmod(remainder, 60)

    return f'{hours:02.0f}:{minutes:02.0f}:{seconds:02.0f}'


def parse_time(time: int) -> Tuple[float, float, float, float]:
    """
    Parses the given time into days, hours, minutes and seconds.
    Useful for formatting time yourself.

    Parameters
    ----------
    time: :class:`int`
        The time in milliseconds.

    Returns
    -------
    Tuple[:class:`float`, :class:`float`, :class:`float`, :class:`float`]
    """
    days, remainder = divmod(time / 1000, 86400)
    hours, remainder = divmod(remainder, 3600)
    minutes, seconds = divmod(remainder, 60)

    return days, hours, minutes, seconds


def _read_track_common(reader: DataReader) -> Tuple[str, str, int, str, bool, Optional[str]]:
    """
    Reads common fields between v1-3 AudioTracks.

    Returns
    -------
    Tuple[str, str, int, str, bool, Optional[str]]
        A tuple containing (title, author, length, identifier, isStream, uri) fields.
    """
    title = reader.read_utfm()
    author = reader.read_utfm()
    length = reader.read_long()
    identifier = reader.read_utf().decode()
    is_stream = reader.read_boolean()
    uri = reader.read_nullable_utf()
    return (title, author, length, identifier, is_stream, uri)


def _write_track_common(track: Dict[str, Any], writer: DataWriter):
    writer.write_utf(track['title'])
    writer.write_utf(track['author'])
    writer.write_long(track['length'])
    writer.write_utf(track['identifier'])
    writer.write_boolean(track['isStream'])
    writer.write_nullable_utf(track['uri'])


def decode_track(track: str,  # pylint: disable=R0914
                 source_decoders: Mapping[str, Callable[[DataReader], Mapping[str, Any]]] = MISSING) -> AudioTrack:
    """
    Decodes a base64 track string into an AudioTrack object.

    Parameters
    ----------
    track: :class:`str`
        The base64 track string.
    source_decoders: Mapping[:class:`str`, Callable[[:class:`DataReader`], Dict[:class:`str`, Any]]]
        A mapping of source-specific decoders to use.
        Some Lavaplayer sources have additional fields encoded on a per-source manager basis, so you can
        specify a mapping of decoders that will handle decoding these additional fields. You can find some
        example decoders within the ``source_decoders`` file. This isn't required for all sources, so ensure
        that you need them before specifying.

        To overwrite library-provided decoders, just specify them within the mapping and the new decoders will
        be used.

    Returns
    -------
    :class:`AudioTrack`
    """
    decoders = DEFAULT_DECODER_MAPPING.copy()

    if source_decoders is not MISSING:
        decoders.update(source_decoders)

    reader = DataReader(track)

    flags = (reader.read_int() & 0xC0000000) >> 30
    version, = struct.unpack('B', reader.read_byte()) if flags & 1 != 0 else (1,)

    title, author, length, identifier, is_stream, uri = _read_track_common(reader)
    extra_fields = {}

    if version == 3:
        extra_fields['artworkUrl'] = reader.read_nullable_utf()
        extra_fields['isrc'] = reader.read_nullable_utf()

    source = reader.read_utf().decode()
    source_specific_fields = {}

    if source in decoders:
        source_specific_fields.update(decoders[source](reader))

    position = reader.read_long()

    track_object = {
        'encoded': track,
        'info': {
            'title': title,
            'author': author,
            'length': length,
            'identifier': identifier,
            'isStream': is_stream,
            'uri': uri,
            'isSeekable': not is_stream,
            'sourceName': source,
            **extra_fields
        }
    }

    return AudioTrack(track_object, 0, position=position, encoder_version=version,
                      source_specific=source_specific_fields)


def encode_track(track: Dict[str, Any],
                 source_encoders: Mapping[str, Callable[[DataWriter, Dict[str, Any]], None]] = MISSING) -> Tuple[int, str]:
    """
    Encodes a track dict into a base64 string, readable by the Lavalink server.

    A track should have *at least* the following keys:
    ``title``, ``author``, ``length``, ``identifier``, ``isStream``, ``uri``, ``sourceName`` and ``position``.

    If the track is a v3 track, it should have the following additional fields:
    ``artworkUrl`` and ``isrc``. isrc can be ``None`` if not applicable.

    Parameters
    ----------
    track: Dict[str, Union[Optional[str], int, bool]]
        The track dict to serialize.
    source_encoders: Mapping[:class:`str`, Callable[[:class:`DataWriter`]]
        A mapping of source-specific encoders to use.
        Some Lavaplayer sources have additional fields encoded on a per-source manager basis, so you can
        specify a mapping of encoders that will handle encoding these additional fields. This isn't required
        for all sources, so ensure that you need them before specifying.

        The mapping must be in the format of something like ``{'http': http_encoder_function}``, where the
        key ``str`` is the name of the source. These functions will only be called if track's ``sourceName``
        field matches.

    Raises
    ------
    :class:`InvalidTrack`
        If the track has unexpected, or missing keys, possibly due to an incompatible version or another reason.

    Returns
    -------
    Tuple[int, str]
        A tuple containing (track_version, encoded_track).
        For example, if a track was encoded as version 3, the return value will be ``(3, '...really long track string...')``.
    """
    track_keys = track.keys()  # set(track) is faster for larger collections, but slower for smaller.

    if not V2_KEYSET <= track_keys:  # V2_KEYSET contains the minimum number of fields required to successfully encode a track.
        missing_keys = [k for k in V2_KEYSET if k not in track]

        raise InvalidTrack(f'Track object is missing keys required for serialization: {", ".join(missing_keys)}')

    if V3_KEYSET <= track_keys:
        return (3, encode_track_v3(track, source_encoders))

    return (2, encode_track_v2(track, source_encoders))


def encode_track_v2(track: Dict[str, Any],
                    source_encoders: Mapping[str, Callable[[DataWriter, Dict[str, Any]], None]] = MISSING) -> str:
    assert V2_KEYSET <= track.keys()

    writer = DataWriter()

    version = struct.pack('B', 2)
    writer.write_byte(version)
    _write_track_common(track, writer)
    writer.write_utf(track['sourceName'])

    if source_encoders is not MISSING and track['sourceName'] in source_encoders:
        source_encoders[track['sourceName']](writer, track)

    writer.write_long(track['position'])

    enc = writer.finish()
    return b64encode(enc).decode()


def encode_track_v3(track: Dict[str, Any],
                    source_encoders: Mapping[str, Callable[[DataWriter, Dict[str, Any]], None]] = MISSING) -> str:
    assert V3_KEYSET <= track.keys()

    writer = DataWriter()
    version = struct.pack('B', 3)
    writer.write_byte(version)
    _write_track_common(track, writer)
    writer.write_nullable_utf(track['artworkUrl'])
    writer.write_nullable_utf(track['isrc'])
    writer.write_utf(track['sourceName'])

    if source_encoders is not MISSING and track['sourceName'] in source_encoders:
        source_encoders[track['sourceName']](writer, track)

    writer.write_long(track['position'])

    enc = writer.finish()
    return b64encode(enc).decode()



"""
MIT License

Copyright (c) 2017-present Devoxin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This module serves to contain all entities which are deserialized using responses from
the Lavalink server.
"""
from enum import Enum as _Enum
from typing import (TYPE_CHECKING, Any, Dict, Final, Literal, Optional, Sequence,
                    Type, TypedDict, TypeVar, Union, cast)

from .errors import InvalidTrack

if TYPE_CHECKING:
    from .abc import DeferredAudioTrack

__all__ = (
    "AudioTrack",
    "EndReason",
    "LoadType",
    "Severity",
    "PlaylistInfo",
    "LoadResultError",
    "LoadResult",
    "Plugin",
    "RawPlugin",
    "RawPlayerState",
    "RawPlayerVoiceState",
    "RawTrackInfo",
    "RawTrack",
    "RawPlayer",
    "RawIpBlock",
    "RawFailingAddress",
    "RawRouteplannerDetails",
    "RawVersion",
    "RawGit",
    "RawInfo",
    "RawMemory",
    "RawCpu",
    "RawStats",
    "RawSession",
    "RawRouteplanner",
)

EnumT = TypeVar('EnumT', bound='Enum')


class Enum(_Enum):
    def __eq__(self, other):
        if self.__class__ is other.__class__:
            return self.value == other.value

        if isinstance(other, str):
            return self.value.lower() == other.lower()

        return False

    @classmethod
    def from_str(cls: Type[EnumT], other: str) -> EnumT:
        try:
            return cls[other.upper()]
        except KeyError:
            try:
                return cls(other)
            except ValueError as error:
                raise ValueError(f'{other} is not a valid {cls.__name__} enum!') from error

    def __str__(self):
        return self.value


class AudioTrack:
    """
    .. _ISRC: https://en.wikipedia.org/wiki/International_Standard_Recording_Code

    Represents an AudioTrack.

    Parameters
    ----------
    data: Union[Dict[str, Union[Optional[str], bool, int]], :class:`AudioTrack`]
        The data to initialise an AudioTrack from.
    requester: :class:`any`
        The requester of the track.
    extra: Dict[Any, Any]
        Any extra information to store in this AudioTrack.

    Attributes
    ----------
    track: Optional[:class:`str`]
        The base64-encoded string representing a Lavalink-readable AudioTrack.
        This is marked optional as it could be ``None`` when it's not set by a custom :class:`Source`,
        which is expected behaviour when the subclass is a :class:`DeferredAudioTrack`.
    identifier: :class:`str`
        The track's id. For example, a youtube track's identifier will look like ``dQw4w9WgXcQ``.
    is_seekable: :class:`bool`
        Whether the track supports seeking.
    author: :class:`str`
        The track's uploader.
    duration: :class:`int`
        The duration of the track, in milliseconds.
    is_stream: :class:`bool`
        Whether the track is a live-stream.
    title: :class:`str`
        The title of the track.
    uri: :class:`str`
        The full URL of track.
    artwork_url: Optional[:class:`str`]
        A URL pointing to the track's artwork, if applicable.
    isrc: Optional[:class:`str`]
        The `ISRC`_ for the track, if applicable.
    position: :class:`int`
        The playback position of the track, in milliseconds.
        This is a read-only property; setting it won't have any effect.
    source_name: :class:`str`
        The name of the source that this track was created by.
    requester: :class:`int`
        The ID of the user that requested this track.
    plugin_info: Optional[Dict[str, Any]]
        Addition track info provided by plugins.
    user_data: Optional[Dict[str, Any]]
        The user data that was attached to the track, if any.
        Any data contained within this dict will be sent to the server when the track instance
        is passed to :func:`BasePlayer.play_track`. This can be useful for, e.g., setting a playback token per-track.
    extra: Dict[str, Any]
        Any extra properties given to this AudioTrack will be stored here.
    """
    __slots__ = ('raw', 'track', 'identifier', 'is_seekable', 'author', 'duration', 'is_stream', 'title', 'uri',
                 'artwork_url', 'isrc', 'position', 'source_name', 'plugin_info', 'user_data', 'extra')

    def __init__(self, data: Union['AudioTrack', Dict[str, Union[Optional[str], bool, int]]], requester: int = 0, **extra):
        if isinstance(data, AudioTrack):
            requester = requester or cast(AudioTrack, data).requester
            extra = {**data.extra, **extra}
            data = data.raw

        self.raw: Dict[str, Union[Optional[str], bool, int]] = data
        info = data.get('info', data)

        try:
            self.track: Optional[str] = data.get('encoded')  # type: ignore
            self.identifier: str = info['identifier']  # type: ignore
            self.is_seekable: bool = info['isSeekable']  # type: ignore
            self.author: str = info['author']  # type: ignore
            self.duration: int = info['length']  # type: ignore
            self.is_stream: bool = info['isStream']  # type: ignore
            self.title: str = info['title']  # type: ignore
            self.uri: str = info['uri']  # type: ignore
            self.artwork_url: Optional[str] = info.get('artworkUrl')  # type: ignore
            self.isrc: Optional[str] = info.get('isrc')  # type: ignore
            self.position: int = info.get('position', 0)  # type: ignore
            self.source_name: str = info.get('sourceName', 'unknown')  # type: ignore
            self.plugin_info: Optional[Dict[str, Any]] = data.get('pluginInfo')  # type: ignore
            self.user_data: Optional[Dict[str, Any]] = data.get('userData')  # type: ignore
            self.extra: Dict[str, Any] = {**extra, 'requester': requester}
        except KeyError as error:
            raise InvalidTrack(f'Cannot build a track from partial data! (Missing key: {error.args[0]})') from error

    def __getitem__(self, name):
        if name == 'info':
            return self

        return super().__getattribute__(name)

    @classmethod
    def from_dict(cls, mapping: dict):
        return cls(mapping)

    @property
    def stream(self) -> bool:
        """
        Property indicating whether this track is a stream.

        .. deprecated:: 5.3.0
            To be consistent with attribute naming, this property has been deprecated
            in favour of ``is_stream``.
        """
        return self.is_stream

    @property
    def requester(self) -> int:
        return self.extra['requester']

    @requester.setter
    def requester(self, requester):
        self.extra['requester'] = requester

    def __repr__(self):
        return f'<AudioTrack title={self.title} identifier={self.identifier}>'


class EndReason(Enum):
    FINISHED = 'finished'
    LOAD_FAILED = 'loadFailed'
    STOPPED = 'stopped'
    REPLACED = 'replaced'
    CLEANUP = 'cleanup'

    def may_start_next(self) -> bool:
        """
        Returns whether the next track may be started from this event.

        This is mostly used as a hint to determine whether the ``track_end_event`` should be
        responsible for playing the next track.

        Returns
        -------
        :class:`bool`
            Whether the next track may be started.
        """
        return self is EndReason.FINISHED or self is EndReason.LOAD_FAILED


class LoadType(Enum):
    TRACK = 'track'
    PLAYLIST = 'playlist'
    SEARCH = 'search'
    EMPTY = 'empty'
    ERROR = 'error'


class Severity(Enum):
    COMMON = 'common'
    SUSPICIOUS = 'suspicious'
    FAULT = 'fault'


class PlaylistInfo:
    """
    Attributes
    ----------
    name: :class:`str`
        The name of the playlist.
    selected_track: :class:`int`
        The index of the selected/highlighted track.
        This will be -1 if there is no selected track.
    """
    __slots__ = ('name', 'selected_track')

    def __init__(self, name: str, selected_track: int = -1):
        self.name: Final[str] = name
        self.selected_track: Final[int] = selected_track

    @classmethod
    def from_dict(cls, mapping: Dict[str, Any]):
        return cls(mapping['name'], mapping.get('selectedTrack', -1))

    @classmethod
    def none(cls):
        return cls('', -1)

    def __repr__(self):
        return f'<PlaylistInfo name={self.name} selected_track={self.selected_track}>'


class LoadResultError:
    """
    Attributes
    ----------
    message: :class:`str`
        The error message.
    severity: :enum:`Severity`
        The severity of the error.
    cause: :class:`str`
        The cause of the error.
    cause_stacktrace: :class:`str`
        The stacktrace for the error.
        May be empty if the server has not been updated to support this.
    """
    __slots__ = ('message', 'severity', 'cause', 'cause_stacktrace')

    def __init__(self, error: Dict[str, Any]):
        self.message: Final[str] = error['message']
        severity = error['severity']
        self.severity: Final[Severity] = severity if isinstance(severity, Severity) else Severity.from_str(str(severity))
        self.cause: Final[str] = error['cause']
        self.cause_stacktrace: Final[str] = error.get('causeStackTrace', '')

    @classmethod
    def create(cls, message: str, severity: Severity, cause: str, stacktrace: str = '') -> 'LoadResultError':
        return cls({'message': message, 'severity': severity, 'cause': cause, 'causeStackTrace': stacktrace})

    def __str__(self):
        return f'{self.message}: {self.cause} ({self.severity})'

    def __repr__(self):
        return f'<LoadResultError severity={self.severity} message={self.message} cause={self.cause}>'


class LoadResult:
    """
    Attributes
    ----------
    load_type: :class:`LoadType`
        The load type of this result.
    tracks: Sequence[Union[:class:`AudioTrack`, :class:`DeferredAudioTrack`]]
        The tracks in this result.
    playlist_info: :class:`PlaylistInfo`
        The playlist metadata for this result.
        The :class:`PlaylistInfo` could contain empty/false data if the :class:`LoadType`
        is not :enum:`LoadType.PLAYLIST`.
    plugin_info: Optional[Dict[:class:`str`, Any]]
        Additional playlist info provided by plugins.
    error: Optional[:class:`LoadResultError`]
        The error associated with this ``LoadResult``.
        This will be ``None`` if :attr:`load_type` is not :attr:`LoadType.ERROR`.
    """
    __slots__ = ('load_type', 'playlist_info', 'tracks', 'plugin_info', 'error')

    def __init__(self, load_type: LoadType, tracks: Sequence[Union[AudioTrack, 'DeferredAudioTrack']],
                 playlist_info: Optional[PlaylistInfo] = None, plugin_info: Optional[Dict[str, Any]] = None,
                 error: Optional[LoadResultError] = None):
        self.load_type: Final[LoadType] = load_type
        self.playlist_info: Final[PlaylistInfo] = playlist_info or PlaylistInfo.none()
        self.tracks: Final[Sequence[Union[AudioTrack, 'DeferredAudioTrack']]] = tracks
        self.plugin_info: Final[Optional[Dict[str, Any]]] = plugin_info
        self.error: Final[Optional[LoadResultError]] = error

    @classmethod
    def empty(cls):
        return LoadResult(LoadType.EMPTY, [])

    @classmethod
    def from_search(cls, tracks: Sequence[Union[AudioTrack, 'DeferredAudioTrack']], playlist_info: Optional[PlaylistInfo] = None,
                    plugin_info: Optional[Dict[str, Any]] = None) -> 'LoadResult':
        return cls(LoadType.SEARCH, tracks, playlist_info, plugin_info)

    @classmethod
    def from_track(cls, track: Union[AudioTrack, 'DeferredAudioTrack'], plugin_info: Optional[Dict[str, Any]] = None) -> 'LoadResult':
        return cls(LoadType.TRACK, [track], plugin_info=plugin_info)

    @classmethod
    def from_playlist(cls, tracks: Sequence[Union[AudioTrack, 'DeferredAudioTrack']], playlist_info: Optional[PlaylistInfo] = None,
                      plugin_info: Optional[Dict[str, Any]] = None) -> 'LoadResult':
        return cls(LoadType.PLAYLIST, tracks, playlist_info, plugin_info)

    @classmethod
    def from_error(cls, error: LoadResultError) -> 'LoadResult':
        return cls(LoadType.ERROR, [], error=error)

    @classmethod
    def from_dict(cls, mapping: dict):
        plugin_info: Optional[dict] = None
        playlist_info: Optional[PlaylistInfo] = PlaylistInfo.none()
        tracks: Sequence[Union[AudioTrack, 'DeferredAudioTrack']] = []

        data: Union[Sequence[Dict[str, Any]], Dict[str, Any]] = mapping['data']
        load_type = LoadType.from_str(mapping['loadType'])

        if isinstance(data, dict):
            plugin_info = data.get('pluginInfo')

        if load_type == LoadType.TRACK:
            tracks = [AudioTrack(data, 0)]  # type: ignore
        elif load_type == LoadType.PLAYLIST:
            playlist_info = PlaylistInfo.from_dict(data['info'])  # type: ignore
            tracks = [AudioTrack(track, 0) for track in data['tracks']]  # type: ignore
        elif load_type == LoadType.SEARCH:
            tracks = [AudioTrack(track, 0) for track in data]  # type: ignore
        elif load_type == LoadType.ERROR:
            error = LoadResultError(data)  # type: ignore
            return cls(load_type, [], playlist_info, plugin_info, error)

        return cls(load_type, tracks, playlist_info, plugin_info)

    @property
    def selected_track(self) -> Optional[AudioTrack]:
        """
        Convenience method for returning the selected track using
        :attr:`PlaylistInfo.selected_track`.

        This could be ``None`` if :attr:`playlist_info` is ``None``,
        or :attr:`PlaylistInfo.selected_track` is an invalid number.

        Returns
        -------
        Optional[:class:`AudioTrack`]
        """
        if self.playlist_info is not None:
            index = self.playlist_info.selected_track

            if 0 <= index < len(self.tracks):
                return self.tracks[index]

        return None

    def __repr__(self):
        return f'<LoadResult load_type={self.load_type} playlist_info={self.playlist_info} tracks=[{len(self.tracks)} item(s)]>'


class Plugin:
    """
    Represents a Lavalink server plugin.

    Parameters
    ----------
    data: Dict[str, Any]
        The data to initialise a Plugin from.

    Attributes
    ----------
    name: :class:`str`
        The name of the plugin.
    version: :class:`str`
        The version of the plugin.
    """
    __slots__ = ('name', 'version')

    def __init__(self, data: Dict[str, Any]):
        self.name: Final[str] = data['name']
        self.version: Final[str] = data['version']

    def __str__(self):
        return f'{self.name} v{self.version}'

    def __repr__(self):
        return f'<Plugin name={self.name} version={self.version}>'


class RawPlugin(TypedDict):
    name: str
    version: str


class RawPlayerState(TypedDict):
    time: int
    position: int
    connected: bool
    ping: int


class RawPlayerVoiceState(TypedDict):
    token: str
    endpoint: str
    sessionId: str
    channelId: Optional[str]


class RawTrackInfo(TypedDict):
    identifier: str
    isSeekable: bool
    author: str
    length: int
    isStream: bool
    position: int
    title: str
    uri: Optional[str]
    artworkUrl: Optional[str]
    isrc: Optional[str]
    sourceName: str


class RawTrack(TypedDict):
    encoded: str
    info: RawTrackInfo
    pluginInfo: Dict[str, Any]
    userData: Dict[Any, Any]


class RawPlayer(TypedDict):
    guildId: str
    track: Optional[RawTrack]
    volume: int
    paused: bool
    state: RawPlayerState
    voice: RawPlayerVoiceState
    filters: Dict[str, Any]  # TODO, however the keys are NotRequired (>=3.11)


class RawIpBlock(TypedDict):
    type: Literal['Inet4Address', 'Inet6Address']
    size: str


class RawFailingAddress(TypedDict):
    failingAddress: str
    failingTimestamp: int
    failingTime: str


class RawRouteplannerDetails(TypedDict):
    ipBlock: RawIpBlock
    failingAddresses: Sequence[RawFailingAddress]
    rotateIndex: str
    ipIndex: str
    currentAddress: str
    currentAddressIndex: str
    blockIndex: str


class RawVersion(TypedDict):
    semver: str
    major: int
    minor: int
    patch: int
    preRelease: Optional[str]
    build: Optional[str]


class RawGit(TypedDict):
    branch: str
    commit: str
    commitTime: int


class RawInfo(TypedDict):
    version: RawVersion
    buildTime: int
    git: RawGit
    jvm: str
    lavaplayer: str
    sourceManagers: Sequence[str]
    filters: Sequence[str]
    plugins: Sequence[RawPlugin]


class RawMemory(TypedDict):
    free: int
    used: int
    allocated: int
    reservable: int


class RawCpu(TypedDict):
    cores: int
    systemLoad: float
    lavalinkLoad: float


class RawStats(TypedDict):
    players: int
    playingPlayers: int
    uptime: int
    memory: RawMemory
    cpu: RawCpu
    frameStats: Literal[None]


class RawSession(TypedDict):
    resuming: bool
    timeout: int


RawRouteplanner = TypedDict('RawRouteplanner', {'class': str, 'details': RawRouteplannerDetails})



import re

from setuptools import setup


version = ''
with open('lavalink/__init__.py') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Version is not set')

setup(
    name='lavalink',
    packages=['lavalink'],
    version=version,
    description='A Lavalink WebSocket & API wrapper built around coverage, reliability and performance.',
    author='Devoxin',
    author_email='luke@serux.pro',
    entry_points={'console_scripts': ['lavalink = lavalink.__main__:main']},
    url='https://github.com/Devoxin/Lavalink.py',
    download_url='https://github.com/Devoxin/Lavalink.py/archive/{}.tar.gz'.format(version),
    keywords=['lavalink'],
    include_package_data=True,
    install_requires=['aiohttp>=3.9.0,<4'],
    extras_require={'docs': ['sphinx',
                             'pygments',
                             'guzzle_sphinx_theme',
                             'enum_tools',
                             'sphinx_toolbox'],
                    'development': ['pylint',
                                    'flake8']}
)
